/*

	Copyright  2003 Frank Condello
	
	Portions Copyright  2001 Dair Grant <mailto:dair@webthing.net>
	and Copyright  2003 Joe Strout <joe@strout.net> (from "LoadSaveJPEG")
	
	"GetData" based on code by Will Leshner <leshner@ljug.com>
	
	Linux port contributed by Jonathan Johnson <jonj@realsoftware.com>

*/

/* PNGu.cpp: Platform neutral functions */

//#define PNG_NO_ASSEMBLER_CODE
//#define PNG_NO_MMX_CODE

#if TARGET_OS_MAC
#define kScreenGamma 1.8
#else
#define kScreenGamma 2.2
#endif

#include "PNGu.h"
#include "png.h"

#if defined(__MWERKS__) && TARGET_API_MAC_CARBON
#include "MachoGlue.h"
#endif

#define kWriteBufferDefaultSize 10240


/* Dynamic Access */
typedef void (*DrawPicturePtr)(REALgraphics, REALpicture, RBinteger, RBinteger, RBinteger, RBinteger, RBinteger, RBinteger, RBinteger, RBinteger);
static void _Graphics_DrawPicture(REALgraphics dest, REALpicture pic, RBinteger x, RBinteger y, RBinteger dw, RBinteger dh, RBinteger sx, RBinteger sy, RBinteger sw, RBinteger sh);
typedef void (*FillRectPtr)(REALgraphics, RBinteger, RBinteger, RBinteger, RBinteger);
static void _Graphics_FillRect(REALgraphics dest, RBinteger x, RBinteger y, RBinteger width, RBinteger height);


/* Local Prototypes */
static REALpicture PNGuLoadPNG(REALfolderItem theFile, RBsingle gamma);
static REALpicture PNGuLoadMaskPNG(REALfolderItem theFile, RBsingle gamma);
static REALpicture _PNGuLoadPNG(REALfolderItem theFile, REALstring theData, RBsingle gamma, RBbool fillMask);
static RBboolRet PNGuSavePNG(REALfolderItem theFile, REALpicture thePicture, RBsingle gamma);
static RBboolRet _PNGuSavePNG(REALfolderItem theFile, REALpicture thePicture, REALstring *data, RBsingle gamma);
static REALpicture PNGuLoadPNGdata(REALstring PNGdata, RBsingle gamma);
static REALpicture PNGuLoadMaskPNGdata(REALstring PNGdata, RBsingle gamma);
static REALstring PNGuSavePNGdata(REALpicture thePicture, RBsingle gamma);
static REALpicture _BuildNewPicture(REALpicture rgbPic, REALpicture maskPic);
static RBboolRet PictureExtHasMask(REALpicture pic);


/* Picture extensions */
REALmethodDefinition PictureExtMethods[] = {
	{ (REALproc)PictureExtHasMask, REALnoImplementation, "HasMask(Extends p As Picture) As Boolean" }
};


/* Module methods */
REALmethodDefinition PNGuModuleMethods[] = {
	{ (REALproc)PNGuLoadMaskPNG, REALnoImplementation, "OpenAsPictureWithMask(file As FolderItem, gamma As Single = 0.0) As Picture" }, 
	{ (REALproc)PNGuLoadPNG, REALnoImplementation, "OpenAsPicture(file As FolderItem, gamma As Single = 0.0) As Picture" }, 
	{ (REALproc)PNGuSavePNG, REALnoImplementation, "SavePicture(file As FolderItem, image As Picture, gamma As Single = 0.0) As Boolean" },
	{ (REALproc)PNGuLoadMaskPNGdata, REALnoImplementation, "GetPictureWithMask(PNGdata As String, gamma As Single = 0.0) As Picture" }, 
	{ (REALproc)PNGuLoadPNGdata, REALnoImplementation, "GetPicture(PNGdata As String, gamma As Single = 0.0) As Picture" },
	{ (REALproc)PNGuSavePNGdata, REALnoImplementation, "GetData(image As Picture, gamma As Single = 0.0) As String" }
};


/* PNGu Module */
REALmoduleDefinition PNGuModule = {
	kCurrentREALControlVersion,
	"PNGu",
	PNGuModuleMethods,
	SizeOfMethods(PNGuModuleMethods)
};


/* Structure to hold write-to-memory information */
typedef struct
{
	png_bytep	buffer;
	UInt32		size;
	UInt32		pos;
}
PNGuWriteBuffer;


#pragma mark -


/* ---------------------------------------------------------------------------
	_PNGuWritefn
	
	libpng write callback - writes to a memory buffer
---------------------------------------------------------------------------- */
static void _PNGuWritefn(png_structp png_ptr, png_bytep data, png_uint_32 length)
{
	PNGuWriteBuffer *saveData = (PNGuWriteBuffer *)png_get_io_ptr(png_ptr);
	png_bytep		newbuffer;
	UInt32			newsize;
		
	if (saveData && saveData->buffer)
	{
		if (saveData->pos + length >= saveData->size)
		{
			newsize = 2 * saveData->size;
			newbuffer = (png_bytep)malloc(newsize);
			
			if (newbuffer == NULL)
			{
				png_error(png_ptr, "PNGu: Write buffer out of memory.");
				return;
			}
			
			memcpy(newbuffer, saveData->buffer, saveData->pos);
			free(saveData->buffer);
			saveData->buffer = newbuffer;
			saveData->size = newsize;
		}

		memcpy(saveData->buffer+saveData->pos, data, length);
		saveData->pos += length;
	}
	else
		png_error(png_ptr, "PNGu: Write buffer error.");
}


/* ---------------------------------------------------------------------------
	_PNGuFlushfn
	
	libpng flush callback - Just a dummy flush...
---------------------------------------------------------------------------- */
static void _PNGuFlushfn(png_structp png_ptr)
{
#pragma unused (png_ptr)
}


/* ---------------------------------------------------------------------------
	_PNGureadfn
	
	libpng read callback - reads from memory buffer
---------------------------------------------------------------------------- */
static void _PNGuReadfn (png_structp png_ptr, png_bytep data, png_uint_32 length)
{
	if (png_ptr->io_ptr == NULL)
	{
		png_error(png_ptr, "PNGu: Read buffer error.");
		return;
	}
	
	memcpy (data, png_ptr->io_ptr, length);
	png_ptr->io_ptr = (char *)png_ptr->io_ptr + length;// Keep GCC4 happy...
	//(char *)png_ptr->io_ptr += length;
}


#pragma mark -


/* ---------------------------------------------------------------------------
	_PNGuLoadPNG
	
	Loads theFile, or uses PNG data passed as a string in theData.
	The image returned may or may not contain a mask.
---------------------------------------------------------------------------- */
static REALpicture _PNGuLoadPNG(REALfolderItem theFile, REALstring theData, RBsingle gamma, RBbool fillMask)
{	
	REALpicture		rgbPic = NULL, finalPic = NULL, maskPic = NULL;
	FILE			*ansiFile = NULL;
	png_byte 		header[8];
	char			*pngDataPtr = NULL;
	png_structp 	png;
	png_infop   	info;
	png_infop   	endinfo;
	png_bytep   	data = NULL;
	png_bytep  		*row_p = NULL;
	png_uint_32 	width, height;
	png_uint_32 	i, j, rowbytes;
	double			fileGamma;
	int 			depth, color;
	Boolean			trans = false;
	
	width = 0;
	height = 0;

	if (theFile)
	{
		// Open the file for reading
		ansiFile = OpenFILEFromFolderItem(theFile, "rb");
		if (ansiFile == NULL) return false;

		// Grab the header
		fread (header, 1, 8, ansiFile);
	}
	else // String
	{
		if ( (theData == NULL) || (theData->CString() == NULL) || (theData->Length() == 0) )
			return false;
		
		// Grab the header
		REALLockString(theData);
		pngDataPtr = (char *)theData->CString();
		memcpy (header, pngDataPtr, 8);
	}

	// Make sure it's a PNG file
	if (!png_check_sig(header, 8)) goto CleanBail;
		
	// Initialize loaders
	png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	info = png_create_info_struct(png);
	endinfo = png_create_info_struct(png);

	// Bail if something goes wrong
	if (setjmp(png->jmpbuf))
	{
		if (row_p) free(row_p);
		png_destroy_read_struct(&png, &info, &endinfo);
		goto CleanBail;
	}

	// Read in the image data...
	
	if (ansiFile)
	{
		png_init_io(png, ansiFile);
		png_set_sig_bytes(png, 8);
	}
	else // String - have to read it ourselves
		png_set_read_fn(png, (png_voidp)pngDataPtr, (png_rw_ptr)_PNGuReadfn);

	png_read_info(png, info);
	png_get_IHDR(png, info, &width, &height, &depth, &color, NULL, NULL, NULL);
	
	// Expand grayscale PNG's to RGB
	if (color == PNG_COLOR_TYPE_GRAY || color == PNG_COLOR_TYPE_GRAY_ALPHA) {
		png_set_gray_1_2_4_to_8(png);// Expand to 8 bit if required
		png_set_gray_to_rgb(png);
	}
	
	// Expand palletes to RGB
	if (color == PNG_COLOR_TYPE_PALETTE)
		png_set_palette_to_rgb(png);
	
	// Expand tRNS chunks to alpha
	if (png_get_valid(png, info, PNG_INFO_tRNS)) {
		png_set_tRNS_to_alpha(png);
     	trans = true;
     }

	// Set pixel format to 8bit ARGB
	png_set_strip_16(png);
	png_set_swap_alpha(png);

	// User-defined gamma
	if ( gamma > 0.0 )
		png_set_gamma(png, kScreenGamma, 1.0/gamma);
	// Auto gamma
	else if (gamma == 0.0)
	{
		if ( png_get_gAMA(png, info, &fileGamma) )// Embedded
			png_set_gamma(png, kScreenGamma, fileGamma);
		else
			png_set_gamma(png, kScreenGamma, 1.0/kScreenGamma);// Screen/System
	}
	// Negative gamma forces raw RGB (no modification)

	// Fill the image buffer
	png_read_update_info(png, info);
	rowbytes = png_get_rowbytes(png, info);
	data = (png_bytep) malloc(rowbytes*height);
	if (data == NULL) goto CleanBail;

	row_p = (png_bytep *) malloc(sizeof(png_bytep)*height);
	if (row_p == NULL) goto CleanBail;

	for (i = 0; i < height; i++)
		row_p[i] = &data[rowbytes*i];

	// Finish off the PNG file
	png_read_image(png, row_p);
	png_read_end(png, endinfo);
	png_destroy_read_struct(&png, &info, &endinfo);
	free(row_p); row_p = NULL;

	// Convert the image data into a picture (and mask)
	
	if ( (color == PNG_COLOR_TYPE_GRAY) || 
		 (color == PNG_COLOR_TYPE_RGB) ||
		 ((color == PNG_COLOR_TYPE_PALETTE) && !trans)
	   )
	{

		rgbPic = REALBuildPictureFromBuffer(width, height, kRBPixelRGB24, data, rowbytes);
		if (fillMask) maskPic = NULL; 
	}
	else if ( (color == PNG_COLOR_TYPE_GRAY_ALPHA) ||
			  (color == PNG_COLOR_TYPE_RGB_ALPHA) ||
		 	  ((color == PNG_COLOR_TYPE_PALETTE) && trans)
	   		)
	{

		rgbPic = REALBuildPictureFromBuffer(width, height, kRBPixelXRGB32, data, rowbytes);
		
		if (fillMask)
		{
			// Copy the alpha into the RGB channels and invert it to work as an RB mask
			for (i = 0; i < rowbytes*height; i+=4)
				for (j = 1; j < 4; j++)
					data[i+j] = 255-data[i];

			// Create the mask image
			maskPic = REALBuildPictureFromBuffer(width, height, kRBPixelXRGB32, data, rowbytes);
		}
	}
	else
	{

		rgbPic = NULL;
		if (fillMask) maskPic = NULL;
	}
	
	// Might as well clean this up early
	if (data) { free(data); data = NULL; }
	
	// Create a new picture and copy the RGB and mask pictures into it.
	finalPic = _BuildNewPicture(rgbPic, maskPic);

CleanBail:
	if (data) free(data);
	if (ansiFile) fclose(ansiFile);
	if (theData) REALUnlockString(theData);
	if (maskPic) { REALUnlockObject((REALobject)maskPic); }
	if (rgbPic) { REALUnlockObject((REALobject)rgbPic); }

	return finalPic;
}


/* ---------------------------------------------------------------------------
	PNGuLoadPNG
	
	Load from File. Ignore alpha.
---------------------------------------------------------------------------- */
static REALpicture PNGuLoadPNG(REALfolderItem theFile, RBsingle gamma)
{
	if (theFile == NULL) { return NULL; }
	return ( _PNGuLoadPNG(theFile, NULL, gamma, false) );
}


/* ---------------------------------------------------------------------------
	PNGuLoadMaskPNG
	
	Load from File. Inverted alpha into mask.
---------------------------------------------------------------------------- */
static REALpicture PNGuLoadMaskPNG(REALfolderItem theFile, RBsingle gamma)
{	
	if (theFile == NULL) { return NULL; }
	return ( _PNGuLoadPNG(theFile, NULL, gamma, true) );
}


/* ---------------------------------------------------------------------------
	PNGuLoadPNGdata
	
	Load from String. Ignore alpha.
---------------------------------------------------------------------------- */
static REALpicture PNGuLoadPNGdata(REALstring theData, RBsingle gamma)
{
	if (theData == NULL) { return NULL; }
	return ( _PNGuLoadPNG(NULL, theData, gamma, false) );
}


/* ---------------------------------------------------------------------------
	PNGuLoadMaskPNGdata
	
	Load from String. Inverted alpha into mask.
---------------------------------------------------------------------------- */
static REALpicture PNGuLoadMaskPNGdata(REALstring theData, RBsingle gamma)
{
	if (theData == NULL) { return NULL; }
	return ( _PNGuLoadPNG(NULL, theData, gamma, true) );
}


#pragma mark -


/* ---------------------------------------------------------------------------
	_PNGuSavePNG
	
	Saves "thePicture" as a 24/32bit RGB/RGBA PNG either to a file or
	to a memory buffer that is later coreced into a REALbasic string.
---------------------------------------------------------------------------- */
static RBboolRet _PNGuSavePNG(REALfolderItem theFile, REALpicture thePicture, REALstring *data, RBsingle gamma)
{
	REALpicture				theMask = NULL, safePic = NULL;
	FILE    				*ansiFile = NULL;
	UInt8					*buf = NULL;
	UInt32					i, j, w, h, bpc, type;
	png_structp				png;
	png_infop				info;
	UInt8					**row = NULL;
	Boolean					tmpsuccess = false;
	REALpictureDescription	pictureDesc;
	PNGuWriteBuffer			pngOut = {NULL, kWriteBufferDefaultSize, 0};
	
	if (thePicture == NULL) return false;
	
	pictureDesc.pictureData = NULL;
	pictureDesc.width = 0;
	pictureDesc.height = 0;
	
	REALLockPictureDescription(thePicture, &pictureDesc);

	// This *can* happen (devs shouldn't send invalid Picture objects but at least we don't crash if they do)
	if (pictureDesc.pictureData == NULL || pictureDesc.width < 1 || pictureDesc.height < 1)
	{
		REALUnlockPictureDescription(thePicture);
		return false;
	}
	
	w = pictureDesc.width;
	h = pictureDesc.height;
	REALUnlockPictureDescription(thePicture);
	
	// If we've got a mask, save an RGBA PNG
	theMask = REALGetPictureMask(thePicture, false);	
	if (theMask)
	{
		// Do another sanity check on the mask....
		pictureDesc.pictureData = NULL;
		pictureDesc.width = 0;
		pictureDesc.height = 0;
		REALLockPictureDescription(theMask, &pictureDesc);
		if (pictureDesc.pictureData == NULL || pictureDesc.width < 1 || pictureDesc.height < 1)
		{
			bpc = 3;
			type = PNG_COLOR_TYPE_RGB;
			REALUnlockPictureDescription(theMask);
			theMask = NULL;
		}
		else
		{
			bpc = 4;
			type = PNG_COLOR_TYPE_RGBA;
			REALUnlockPictureDescription(theMask);
		}
	}
	else
	{
		bpc = 3;
		type = PNG_COLOR_TYPE_RGB;
	}
	
	// Allocate image buffer and fill it
	buf = (UInt8 *)malloc(w*h*bpc);
	if (buf == NULL) goto CleanBail;
	
	// Create a temp 32 bit picture so BuildImageDataFromPicture doesn't have to deal with oddballs.
	// This is a round-a-bout way to fix crashes on Win32 with non-32-bit Pictures.
	// This also fixes a bug on Linux (which always used dynamic access for BuildNewPicture) 
	// since previously this just sent the masked RGB pic which meant the alpha was destructively
	// applied when saving. BuildNewPicture now expects no masks, on either picture...
	
	if (theMask)
	{
		REALgraphics	g;
		REALpicture		maskcopy;
		long			colsave;
		
		// Copy the mask
		maskcopy = REALNewPicture( w, h, 32 );
		if (!maskcopy) goto CleanBail;
		g = REALGetPictureGraphics( maskcopy );
		if (!g) {
			REALUnlockObject( (REALobject)maskcopy );
			goto CleanBail;
		}
		_Graphics_DrawPicture( g, theMask, 0, 0, w, h, 0, 0, w, h );
		REALUnlockObject( (REALobject)g ); g = NULL;
		
		// Wipe the original mask
		g = REALGetPictureGraphics( theMask );
		if (!g || !REALGetPropValue( (REALobject)g, "ForeColor", &colsave ) ) {
			REALUnlockObject( (REALobject)maskcopy );
			goto CleanBail;
		}
		REALSetPropValue( (REALobject)g, "ForeColor", (long)0x000000 );
		_Graphics_FillRect(g, 0, 0, w, h);
		REALSetPropValue( (REALobject)g, "ForeColor", (long)colsave );
		
		// Build the new picture
		safePic = _BuildNewPicture(thePicture, maskcopy);
		
		// Restore the original mask and clean up
		_Graphics_DrawPicture( g, maskcopy, 0, 0, w, h, 0, 0, w, h );
		REALUnlockObject( (REALobject)g );
		REALUnlockObject( (REALobject)maskcopy );
	}
	else
		safePic = _BuildNewPicture(thePicture, NULL);
	
	if (safePic)
	{
		if (type == PNG_COLOR_TYPE_RGB)
			tmpsuccess = BuildImageDataFromPicture(safePic, w, h, buf);
		else
			tmpsuccess = BuildImageDataFromPictureAndMask(safePic, theMask, w, h, buf);
		
		REALUnlockObject((REALobject)safePic); safePic = NULL;
	}
	
	if (theMask) { REALUnlockObject((REALobject)theMask); theMask = NULL; }

	if (!tmpsuccess) goto CleanBail;
	
	// Allocate row pointers
	row = (UInt8 **)malloc(sizeof(UInt8 *)*h);
	if (row == NULL) goto CleanBail;
	
	if (theFile)
	{
		// Open the file for writing
		ansiFile = OpenFILEFromFolderItem(theFile, "wb");
		if (ansiFile == NULL) goto CleanBail;
	}
	else
	{
		// Initialize a memory buffer
		pngOut.buffer = (png_bytep)malloc(kWriteBufferDefaultSize);
		if (pngOut.buffer == NULL) goto CleanBail;
	}
	
	// Initialize PNG write struct
 	png = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png == NULL) goto CleanBail;
 	
 	// Initialize PNG info struct
	info = png_create_info_struct(png);
	if (info == NULL) {
		png_destroy_write_struct(&png, (png_infopp)NULL);
		goto CleanBail;
	}
   
	// Clean up if pnglib bails
	if (setjmp(png->jmpbuf)) {
		png_destroy_write_struct(&png, &info);
		goto CleanBail;
	} 
	
	// Setup file IO - either standard file stream or our custom memory buffer
	if (ansiFile)
		png_init_io(png, ansiFile);
	else
		png_set_write_fn(png, &pngOut, (png_rw_ptr)_PNGuWritefn, (png_flush_ptr)_PNGuFlushfn);
	
	// Set the PNG header
	png_set_IHDR(png, info, w, h, 8, type, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
	
	// Set gamma chunk
	if ( gamma > 0.0 )
		png_set_gAMA(png, info, 1.0/gamma);// User defined
	else if ( gamma == 0.0 )
		png_set_gAMA(png, info, 1.0/kScreenGamma);// System defined
	// -1 ommits gAMA chunk

	// Write the PNG...
	
	png_write_info(png, info);
	
	for (i = 0, j = h-1; i < h; i++, j--)
		row[h-j-1] = buf+i*w*bpc;
	
	png_write_image(png, row);
	png_write_end(png, info);
	
	// Clean up
	png_destroy_write_struct(&png, &info);
	free(row); row = NULL;
	free(buf); buf = NULL;
	if (ansiFile) fclose(ansiFile);
	
	// Copy data into an RB string if writing to memory
	if (pngOut.buffer)
	{
		*data = REALBuildString((char *)pngOut.buffer, pngOut.pos);
		free(pngOut.buffer);	
	}
	
	return true;

CleanBail:
	if (theMask) REALUnlockObject((REALobject)theMask);
	if (row) free(row);
	if (buf) free(buf);
	if (ansiFile) fclose(ansiFile);
	if (pngOut.buffer) free(pngOut.buffer);
	return false;
}


/* ---------------------------------------------------------------------------
	PNGuSavePNG
	
	Writes a PNG file. Returns false on failure.
---------------------------------------------------------------------------- */
static RBboolRet PNGuSavePNG(REALfolderItem theFile, REALpicture thePicture, RBsingle gamma)
{
	if (theFile == NULL || thePicture == NULL) { return false; }
	return ( _PNGuSavePNG(theFile, thePicture, NULL, gamma) );
}


/* ---------------------------------------------------------------------------
	PNGuSavePNGdata
	
	Returns a string containing PNG data.
---------------------------------------------------------------------------- */
static REALstring PNGuSavePNGdata(REALpicture thePicture, RBsingle gamma)
{
	REALstring pngData = NULL;
	
	if (thePicture == NULL) { return NULL; }
	
	if ( _PNGuSavePNG(NULL, thePicture, &pngData, gamma) )
		return ( pngData );
		
	return NULL;
}


#pragma mark -


/* ---------------------------------------------------------------------------
	PictureExtHasMask
	
	Accessing a mask in RB creates it, so Picture.Mask = Nil is always false.
	This adds a missing bit of functionality, and is useful when used
	in conjunction with PNGu.OpenAsPictureWithMask to determine whether a
	PNG file had an alpha channel or 1-bit transparency index.
---------------------------------------------------------------------------- */
static RBboolRet PictureExtHasMask(REALpicture pic)
{
	REALpicture				theMask = NULL;
	
	if (pic != NULL)
	{
		theMask = REALGetPictureMask(pic, false);	
		if (theMask)
		{
			REALUnlockObject((REALobject)theMask);
			return true;
		}
	}
	
	return false;
}


#pragma mark -


/* ---------------------------------------------------------------------------
	_Graphics_DrawPicture
	
	Dynamic access wrapper for Graphics.DrawPicture
---------------------------------------------------------------------------- */
static void _Graphics_DrawPicture(REALgraphics dest, REALpicture pic, RBinteger x, RBinteger y, RBinteger dw, RBinteger dh, RBinteger sx, RBinteger sy, RBinteger sw, RBinteger sh)
{
	DrawPicturePtr DrawPicture = NULL;

	if (dest && pic)
	{
		DrawPicture = (DrawPicturePtr)REALLoadObjectMethod( (REALobject)dest, 
					  "DrawPicture( p as Picture, x as Integer, y as Integer, dw as Integer, dh as Integer, sx as Integer, sy as Integer, sw as Integer, sh as Integer )" );
		if (DrawPicture) DrawPicture( dest, pic, x, y, dw, dh, sx, sy, sw, sh );
	}
}


/* ---------------------------------------------------------------------------
	_Graphics_FillRect
	
	Dynamic access wrapper for Graphics.FillRect
---------------------------------------------------------------------------- */
static void _Graphics_FillRect(REALgraphics dest, RBinteger x, RBinteger y, RBinteger width, RBinteger height)
{
	FillRectPtr FillRect = NULL;

	if (dest)
	{
		FillRect = (FillRectPtr)REALLoadObjectMethod( (REALobject)dest, 
				   "FillRect( x as Integer, y as Integer, width as Integer, height as Integer )" );
		if (FillRect) FillRect( dest, x, y, width, height );
	}
}


/* ---------------------------------------------------------------------------
	_BuildNewPicture
	
	Creates a new picture from separate RGB and mask pictures.

	IMPORTANT! For proper results both rgbPic and maskPic should have solid
	masks (e.g. no transparency whatsoever) 
---------------------------------------------------------------------------- */
REALpicture _BuildNewPicture(REALpicture rgbPic, REALpicture maskPic)
{
	unsigned long width;
	unsigned long height;
	REALpictureDescription	pictureDesc;
	REALpicture newPicture = NULL;
	
	if (rgbPic)
	{
		REALLockPictureDescription( rgbPic, &pictureDesc );
		width = pictureDesc.width;
		height = pictureDesc.height;
		REALUnlockPictureDescription( rgbPic );
		
		REALpicture mask;
		REALgraphics dest;

		newPicture = REALNewPicture( width, height, 32 );
		
		if (!newPicture) return NULL;

		dest = REALGetPictureGraphics( newPicture );
		
		if (!dest)
		{
			REALUnlockObject( (REALobject)newPicture );
			return NULL;
		}
		
		// Draw the picture
		_Graphics_DrawPicture( dest, rgbPic, 0, 0, width, height, 0, 0, width, height );
		REALUnlockObject( (REALobject)dest );
		
		if (maskPic)
		{
			// Handle the mask
			mask = REALGetPictureMask( newPicture, true );
			
			if (mask)
			{
				dest = REALGetPictureGraphics( mask );
				
				if (dest)
				{
					_Graphics_DrawPicture( dest, maskPic, 0, 0, width, height, 0, 0, width, height );
					REALUnlockObject( (REALobject)dest );
				}

				REALUnlockObject( (REALobject)mask );
			}
		}
	}
		
	return newPicture;
}


#pragma mark -


/* ---------------------------------------------------------------------------
	PluginEntry
---------------------------------------------------------------------------- */
void PluginEntry(void)
{
	UInt32		n;
	
#if defined(__MWERKS__) && TARGET_API_MAC_CARBON
	// Try to use CoreGraphics in CFM Carbon, but fallback to Quickdraw if it's not
	// available (<10.2). NOTE: Mach-O targets require 10.2 or better...
	machFuncsAvailable = InitMachoFunctionPointers();
#endif

	if (REALGetRBVersion() >= 5.5)
		REALRegisterModule(&PNGuModule);
	
	// Register the global methods
	for (n = 0; n < ( SizeOfMethods(PictureExtMethods) ); n++)
		REALRegisterMethod(&PictureExtMethods[n]);
}

